package SAML::SPMetadata::SP ;

# We bend the variable naming convention in some cases since we want the
# variable name to match the XML attribute name:
## no critic (NamingConventions::Capitalization)
## no critic (CodeLayout::ProhibitParensWithBuiltins)

use strict ;
use warnings ;
use autodie ;

use Data::Dumper ;
use Carp ;
use Text::Wrap ;
use JSON;

use SAML::SPMetadata::X509Certificate ;
use SAML::SPMetadata::KeyDescriptor ;
use SAML::SPMetadata::ContactPerson ;
use SAML::SPMetadata::AssertionConsumerService ;
use SAML::SPMetadata::Utility qw/
  get_node_attribute
  get_elements_matching
  trim
/ ;

use Readonly;
Readonly my $CertLineLength => 64;

# Make this a Moose class
use Moose ;
use namespace::autoclean;


has metadata => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
) ;

has entityID => (
  is        => 'rw',
  isa       => 'Maybe[Str]',
  required  => 1,
    ) ;

has validUntil => (
  is        => 'rw',
  isa       => 'Maybe[Str]',
  required  => 0,
    ) ;

has KeyDescriptors => (
  is        => 'rw',
  isa       => 'ArrayRef[SAML::SPMetadata::KeyDescriptor]',
  required  => 1,
    ) ;

has AssertionConsumerServices => (
  is        => 'rw',
  isa       => 'ArrayRef[SAML::SPMetadata::AssertionConsumerService]',
  required  => 0,
    ) ;

has ContactPersons => (
  is        => 'rw',
  isa       => 'ArrayRef[SAML::SPMetadata::ContactPerson]',
  required  => 1,
    ) ;

sub from_xml_string {
    my ($xmlstring) = @_ ;

    my $parser = XML::LibXML->new();
    my $doc = $parser->parse_balanced_chunk($xmlstring) ;

    my (@nodes) ;

    # 1. Make sure there is a single child node and that it is an
    # EntityDescriptor node.
    my @childnodes = $doc->childNodes();

    my @EntityDescriptor_nodes = () ;
    foreach my $child (@childnodes) {
        if ($child->nodeName() =~ m{EntityDescriptor}xsm) {
           push(@EntityDescriptor_nodes, $child) ;
        }
    }

    if (scalar(@EntityDescriptor_nodes) == 0) {
        croak 'found no EntityDescriptor nodes?!?' ;
    } elsif (scalar(@EntityDescriptor_nodes) > 1) {
        croak 'found more than one EntityDescriptor nodes (should only be one)' ;
    }

    my $EntityDescriptor_node = $EntityDescriptor_nodes[0] ;

    if (!$EntityDescriptor_node) {
        croak 'found no EntityDescriptor elements' ;
    } elsif ($EntityDescriptor_node->nodeName() !~ m{EntityDescriptor}xsm) {
        croak 'root element not an EntityDescriptor element' ;
    }

    # 2. Get the entityID
    my $entityID = get_node_attribute($EntityDescriptor_node, 'entityID') ;

    # 3. Get the validUntil attribute (may not exist)
    my $validUntil = get_node_attribute($EntityDescriptor_node, 'validUntil') ;
    if (!$validUntil) {
        $validUntil = q{} ;
    }

    # 4. Get all the KeyDescriptors elements.
    @nodes = get_elements_matching($EntityDescriptor_node,
                                   'KeyDescriptor') ;

    my @KeyDescriptors = () ;
    foreach my $node (@nodes) {
        # Get the "use" attribute.
        my $usage = get_node_attribute($node, 'use') ;

        # Get all the X509Certificate elements of this KeyDescriptor element.
        my @cert_nodes = $node->findnodes(q{.//*[local-name()='X509Certificate']}) ;
        my @x509_certs = () ;
        foreach my $cert_node (@cert_nodes) {
            my $cert_string = trim($cert_node->textContent()) ;
            my @lines = split(/\n/xsm, $cert_string) ;

            # Make sure $cert_string has lines of length $CertLineLength.
            $Text::Wrap::columns = $CertLineLength ; ## no critic (ProhibitPackageVars);
            $cert_string = wrap(q{}, q{}, @lines);

            # Create a Certificate object.
            my $cert_object;
            eval {
                $cert_object =
                  SAML::SPMetadata::X509Certificate::from_pem($cert_string) ;
            } ;

            if (!$cert_object) {
                ## no critic (ErrorHandling::RequireCarping);
                die 'error parsing certificate';
            }

            push (@x509_certs, $cert_object) ;
        }

        # Create a KeyDescriptor object
        my $keyd_object = SAML::SPMetadata::KeyDescriptor->new(
            X509Certificates => \@x509_certs,
            usage            => $usage,
            ) ;

        push(@KeyDescriptors, $keyd_object) ;
    }

    # 5. Get all the ContactPerson elements.
    @nodes = get_elements_matching($EntityDescriptor_node,
                                   'ContactPerson') ;

    my @ContactPersons = () ;
    foreach my $node (@nodes) {
        my $ContactPerson = SAML::SPMetadata::ContactPerson::from_node($node) ;
        push(@ContactPersons, $ContactPerson) ;
    }

    # 6. Get all the AssertionConsumerServices
    my @AssertionConsumerServices = () ;
    @nodes = get_elements_matching($EntityDescriptor_node,
                                   'AssertionConsumerService') ;
    foreach my $node (@nodes) {
        my $ACS = SAML::SPMetadata::AssertionConsumerService::from_node($node) ;
        push(@AssertionConsumerServices, $ACS) ;
    }


    my $sp = SAML::SPMetadata::SP->new(
        'metadata'       => $xmlstring,
        'entityID'       => $entityID,
        'validUntil'     => $validUntil,
        'KeyDescriptors' => \@KeyDescriptors,
        'ContactPersons' => \@ContactPersons,
        'AssertionConsumerServices' => \@AssertionConsumerServices,
        ) ;


    return $sp ;
}

sub to_json_string {
    my $self = shift ;

    my $structure = {} ;

    $structure->{'entityID'}   = $self->entityID() ;
    $structure->{'validUntil'} = $self->validUntil() ;

    my @kds = () ;
    my @KeyDescriptors = @{ $self->KeyDescriptors() } ;
    foreach my $kd (@KeyDescriptors) {
        push(@kds, $kd->to_hash()) ;
    }
    $structure->{'KeyDescriptors'} = \@kds ;

    my @cps = () ;
    my @ContactPersons = @{ $self->ContactPersons() } ;
    foreach my $cp (@ContactPersons) {
        push(@cps, $cp->to_hash()) ;
    }
    $structure->{'ContactPersons'} = \@cps ;

    my @ACSes = () ;
    my @AssertionConsumerServices = @{ $self->AssertionConsumerServices() } ;
    foreach my $ACS (@AssertionConsumerServices) {
        push(@ACSes, $ACS->to_hash()) ;
    }
    $structure->{'AssertionConsumerServices'} = \@ACSes ;

    return encode_json($structure) ;
}

sub get_bindings_to_locations {
    my $self = shift ;

    my @AssertionConusumerServices = @{ $self->AssertionConsumerServices() } ;

    my %Binding_to_Location = () ;
    foreach my $ACS (@AssertionConusumerServices) {
        my $Binding  = $ACS->Binding() ;
        my $Location = $ACS->Location() ;
        $Binding_to_Location{$Binding} = $Location ;
    }

    return %Binding_to_Location ;
}

1;
