package SAML::SPMetadata::Validate ;

# We bend the variable naming convention in some cases since we want the
# variable name to match the XML attribute name:
## no critic (NamingConventions::Capitalization)
## no critic (Modules::RequireNoMatchVarsWithUseEnglish)
## no critic (CodeLayout::ProhibitParensWithBuiltins)
## no critic (ErrorHandling::RequireCheckingReturnValueOfEval)

use strict ;
use warnings ;
use autodie ;

use base qw( Exporter ) ;
our @EXPORT_OK = qw
    (
     validate
  ) ;

use Carp ;
use English;
use File::Spec ;
use File::Copy ;
use File::Temp ;
use XML::LibXML;  # libxml-libxml-perl
use Text::Wrap ;
use IPC::Run3 ;   # libipc-run3-perl
use Crypt::X509 ; # libcrypt-x509-perl
use MIME::Base64;
use Date::Parse ;
use Data::Dumper ;

use Readonly;
Readonly my $CertLineLength => 64;

use SAML::SPMetadata::SP ;
use SAML::SPMetadata::Utility qw/
  trim
/ ;

my $DEBUG = 0 ;

my $SCHEMA_DIR     = '/usr/share/saml-validate/schema' ;

my %ERROR_MESSAGES = (
  '100' => 'is empty',
  '200' => 'contains <?xml',
  '300' => 'is not valid XML',
  '350' => 'is not valid SAML 2.0 metadata: %%1',
  '370' => 'contains a non-parsable certificate',
  '400' => 'contains no entityID',
  '410' => 'validUntil is in the past',
  '510' => 'contains a certificate (cert%%1) that has expired',
  '610' => 'contains a %%1 element that is empty but should not be',
) ;

sub schema_dir {
    my ($dir) = @_ ;

    if ($dir) {
        $SCHEMA_DIR = $dir ;
    }
    return $SCHEMA_DIR ;
}

sub debug {
    my ($debug) = @_ ;

    if (defined($debug)) {
        $DEBUG = $debug ;
    }
    return $DEBUG ;
}


# Returns the ($sp, $errors_href) where $sp is a SAML::SPMetadata::SP
# object and $errors_href is a reference to a hash containing error
# messages.
sub validate {
    my ($fragment) = @_ ;

    # $sp will be a SAML::SPMetadata::SP object.
    my $sp ;

    # The validation errors.
    my %errors  = () ;

    my @cmd ;
    my ($stdout, $stderr, $exit_value) ;
    my $exit_status = 0 ;

    ## ### #    ## ### #    ## ### #    ## ### #    ## ### #    ## ### #
    my $return_values_fref = sub {
        return ($sp, \%errors) ;
    } ;
    ## ### #    ## ### #    ## ### #    ## ### #    ## ### #    ## ### #
    my $add_error_fref = sub {
        my ($err_code, $text) = @_ ;

        my $error_string = $ERROR_MESSAGES{$err_code} ;

        if ($text) {
            $error_string =~ s{%%1}{$text}xsm ;
        }

        if ($errors{$err_code}) {
            my @cur_array = @{ $errors{$err_code} } ;
            push(@cur_array, $error_string) ;
            $errors{$err_code} = \@cur_array ;
        } else {
            $errors{$err_code} = [ $error_string ] ;
        }
        return ;
    } ;
    ## ### #    ## ### #    ## ### #    ## ### #    ## ### #    ## ### #

    # 0. Is there anything in the string?
    $fragment = trim($fragment) ;
    if (!$fragment) {
        $add_error_fref->('100') ;

        # No need to continue.
        return $return_values_fref->() ;
    }

    # Create a temporary file containing the string.
    my $tmp      = File::Temp->new(UNLINK => 1) ;
    my $tempfile = $tmp->filename() ;
    spew($fragment, $tempfile) ;

    # 1. Does the fragment file contain the string "<?xml"?
    progress('testing for <?xml') ;
    if (index($fragment, '<?xml') >= 0) {
        $add_error_fref->('200') ;
    }

    # 2. Valid XML.
    progress('testing for valid XML') ;
    @cmd = ('xmllint', q{--noout}, $tempfile) ;
    ($stdout, $stderr, $exit_value) = run_command(@cmd) ;
    if ($stderr) {
        $add_error_fref->('300') ;
        return $return_values_fref->() ;
    }

    # 3. Validate using xmllint and schema.
    progress('validating against SAML 2.0 metadata schema') ;
    ($stdout, $stderr, $exit_value) = validate_against_schema($tempfile) ;
    if ($exit_value) {
        # Remove the temporary file name from the error message.
        $stderr =~ s{$tempfile[ ]fails[ ]to[ ]validate}{}xsm ;
        $stderr =~ s{^$tempfile\S+}{}xsm ;
        $stderr = trim($stderr) ;
        $add_error_fref->('350', $stderr) ;
    }

    # If there are any errors at this point, return now?
    # 4. Parse the XML
    eval {
        $sp = SAML::SPMetadata::SP::from_xml_string($fragment) ;
    } ;

    my $aterror = $EVAL_ERROR ;
    if ($aterror =~ m{error[ ]parsing[ ]certificate}ixsm) {
        $add_error_fref->('370') ;
        # No need to continue.
        return $return_values_fref->() ;
    }

    if ($sp) {
        my $entityID = $sp->entityID() ;

        if (!$entityID) {
            $add_error_fref->('400') ;
        }

        # 5. If there is a validUntil string, parse it and make sure it is in
        # the future.
        my $validUntil = $sp->validUntil() ;
        if ($validUntil) {
            # Convert to a DateTime object.
            my $dt = str2time($validUntil);
            if ($dt < time()) {
                $add_error_fref->('410') ;
            }
        }
    }

    # Look for empty elements that shouldn't be empty.
    ## no critic (Variables::ProhibitPackageVars)
    my $parser = XML::LibXML->new();
    $XML::XPath::Namespaces = 0;
    my $doc;
    eval {
        $doc = $parser->parse_balanced_chunk($fragment);
    };
    my $parser_at_error = $EVAL_ERROR;
    # If we get a parser error no reason to look for elements.
    if (!$parser_at_error) {
        ## ### #    ## ### #    ## ### #    ## ### #    ## ### #    ## ### #
        my $empty_elements_fref = sub {
            my ($element_path) = @_ ;
            # Explode path on the "/"
            my @parts = split(m{\/}xsm, $element_path);

            # Now join with the local-name qualifier so we can ignore namespaces
            my @localized_parts = ();
            foreach my $part (@parts) {
                push(@localized_parts, qq%[local-name()='$part']%)
            }
            my $new_path = q{*} . join(q{/*}, @localized_parts);
            # warn $new_path;
            my $result = $doc->findnodes($new_path);
            foreach my $node ($result->get_nodelist) {
                my $node_name    = $node->nodeName;
                my $text_content = trim($node->textContent) ;
                if (!$text_content) {
                    return 1;
                }
            }
            return 0;
        } ;
        ## ### #    ## ### #    ## ### #    ## ### #    ## ### #    ## ### #

        my %these_elements_cannot_be_empty = (
            'ServiceName'             => 'EntityDescriptor/SPSSODescriptor/AttributeConsumingService',
            'ServiceDescription'      => 'EntityDescriptor/SPSSODescriptor/AttributeConsumingService',
            'OrganizationName'        => 'EntityDescriptor/Organization',
            'OrganizationDisplayName' => 'EntityDescriptor/Organization',
            'OrganizationURL'         => 'EntityDescriptor/Organization',
        ) ;

        foreach my $element (sort keys %these_elements_cannot_be_empty) {
            my $prefix = $these_elements_cannot_be_empty{$element};
            my $path = "$prefix/$element";
            if ($empty_elements_fref->($path)) {
                # warn "found empty element $element";
                $add_error_fref->('610', $element) ;
            }
        }
    }

    return $return_values_fref->() ;
}

# Returns a hash with these keys:
#   entityID
#   entityID_domain
#   validUntil
#   is_urn
#   qualifier (e.g., 'md')
#   contact_person: a hash
#      contactType
#      GivenName
#      EmailAddress
#   certs_info_aref (an array of certificate information):
#     [use, expiration_date_epoch, pemstring]
#   assertion_consumer_services (a hash mapping ACS bindings to SP
#     endpoints)
#
# If the entityID is a URN rather than a URL, and if the entityID starts
# with "urn:", then entityID_domain will be set to undef and the is_urn key
# will be set to 1.
sub parse_xml {
    my ($xmlstring) = @_ ;

    if (!$xmlstring) {
        exit_error('XML fragment appears to be empty') ;
    }

    my %results = () ;
    my @attributes ;

    my $parser = XML::LibXML->new();
    my $doc = $parser->parse_balanced_chunk($xmlstring) ;

    my @children = $doc->childNodes ;

    # The base_node will be the EntityDescriptor element.
    my $base_node ;
    foreach my $child (@children) {
        if ($child->nodeName() =~ m{EntityDescriptor}ixsm) {
            $base_node = $child ;
            if ($child->nodeName() =~ m{^(\w+):}xsm) {
                $results{'prefix'} = $1 ;
            }
        }
    }

    @attributes = $base_node->attributes() ;
    foreach my $attribute (@attributes) {
        my $name = $attribute->name() ;
        if ($name eq 'entityID') {
            $results{'entityID'} = $attribute->getValue() ;
        }
    }

    my $entityID_domain ;
    if ($results{'entityID'}) {
        if ($results{'entityID'} =~ m{^(?:http|https)://([^\/]+)(\/)?}xsm) {
            $entityID_domain = $1 ;
            progress("entityID_domain is $entityID_domain") ;
            $results{'entityID_domain'} = $entityID_domain ;
            $results{'is_urn'} = 0 ;
        } elsif ($results{'entityID'} =~ m{^urn:(\S+)$}xsm) {
            $results{'entityID_domain'} = undef ;
            $results{'is_urn'} = 1 ;
        } else {
            $results{'is_urn'} = 0 ;
        }
    }

    my $contact_node ;
    foreach my $child ($base_node->childNodes) {
        if ($child->nodeName() =~ m{ContactPerson}xsm) {
            $contact_node = $child ;
        }
    }

    if ($contact_node) {
        my %contact_person = parse_node_contact($contact_node) ;
        if ($contact_person{'contactType'}) {
            $results{'contactType'} = $contact_person{'contactType'} ;
        }
        $results{'contact_person'} = \%contact_person ;
    }

    # Get any certificates
    my @certs_info = parse_node_certs($xmlstring, $doc) ;
    $results{'certs_info_aref'} = \@certs_info ;

    # Get any AssertionConsumerService elements. These elements are
    # children of the SPSSODescriptor element. Note that SAML 2.0 metadata
    # can have zero, one, ore more than one SPSSODescriptor element
    # (although usually there is just one). Right  now we choose the
    # _last_ such element.
    my $SPSSODescriptor_node ;
    foreach my $child ($base_node->childNodes) {
        if ($child->nodeName() =~ m{SPSSODescriptor}xsm) {
            $SPSSODescriptor_node = $child ;
        }
    }

    # Get all the AssertionConsumerService child elements of
    # $SPSSODescriptor_node.
    my %ACS = SAML::SPMetadata::AssertionConsumerService::get_ACS_from_SPSSODescriptor($SPSSODescriptor_node) ;
    $results{'assertion_consumer_services'} = \%ACS ;

    # Get all empty elements.QQQQ


    return %results ;
}

# Returns a hash with these keys:
#
# EmailAddress
# GivenName
# contactType

sub parse_node_contact {
    my ($contact_node) = @_ ;

    my @attributes = () ;

    my %contact_person = () ;

    if ($contact_node) {
        @attributes = $contact_node->attributes() ;
        foreach my $attribute (@attributes) {
            my $name = $attribute->name() ;
            if ($name eq 'contactType') {
                $contact_person{'contactType'} = $attribute->getValue() ;
            }
        }

        foreach my $node ($contact_node->childNodes()) {
            my $nodeName = $node->nodeName() ;
            if ($nodeName =~ m{EmailAddress}xsm) {
                $contact_person{'EmailAddress'} = $node->textContent() ;
            }
            if ($nodeName =~ m{GivenName}xsm) {
                $contact_person{'GivenName'} = $node->textContent() ;
            }
        }
    }

    return %contact_person ;
}

sub get_certs_from_node {
    my ($node) = @_ ;

    my @cert_nodes ;

    # The following xpath means: "starting at the current node (".") search
    # everywhere ("//") for any element ("*") that has a "local-name"
    # (i.e., the part of their name without the namespace qualifier) equal
    # to "X509Certificate".
    @cert_nodes = $node->findnodes(q{.//*[local-name()='X509Certificate']}) ;

    my @certs = () ;
    foreach my $cert_node (@cert_nodes) {
        my $cert_string = trim($cert_node->textContent()) ;
        my @lines = split(/\n/xsm, $cert_string) ;

        # Make sure $cert_string has lines of length $CertLineLength.
        $Text::Wrap::columns = $CertLineLength ; ## no critic (ProhibitPackageVars);
        $cert_string = wrap(q{}, q{}, @lines);

        push (@certs, $cert_string) ;
    }

    return @certs ;
}

# Return an array of certificate information. Enty array entry contains an
# array reference:
#
# [cert-user, cert-enddate-epoch, cert-pem]
sub parse_node_certs {
    my ($xmlstring, $base_node) = @_ ;

    my @certs_info = () ;

    # We use 'local-name' to ignore troublesome namespaces.
    my @kd_nodes = $base_node->findnodes(q{//*[local-name()='KeyDescriptor']}) ;

    # Iterate over each KeyDescriptor node looking for certificates.
    foreach my $kd_node (@kd_nodes) {
        # First get the "use" of this KeyDescriptor (signing or
        # encryption).
        my $use = get_node_attribute($kd_node, 'use') ;

        # Now parse this node for certificates.
        my @certs = get_certs_from_node($kd_node) ;
        progress('in KeyDescriptor found ' . scalar(@certs) . ' certs.') ;

        foreach my $cert (@certs) {
            # Get the expiration date.
            my $cert_der = decode_base64($cert);

            my $decoded = Crypt::X509->new(cert => $cert_der) ;
            my $enddate_epoch = $decoded->not_after() ;

            push(@certs_info, [$use, $enddate_epoch, $cert]) ;
        }
    }

    return @certs_info ;
}

sub progress {
    ## no critic (InputOutput::RequireCheckedSyscalls)
    my ($msg) = @_ ;

    if ($DEBUG) {
        print "info: $msg\n" ;
    }
    return ;
}



# Validate the SP fragment against the SAML schema
#
# For details on the XML schema that SAML metadata uses see
# http://docs.oasis-open.org/security/saml/v2.0/saml-metadata-2.0-os.pdf
#
# Also, see catalog.template for all the xsd schema. Note that the "ws*
# namespaces are for ADFS.
#
# See also
# https://wiki.shibboleth.net/confluence/display/CONCEPT/MetadataCorrectness

sub validate_against_schema {
    my ($sp_metadata_file) = @_ ;

    #  my $MD_SCHEMA=schema/saml-schema-metadata-2.0.xsd
    #WSFED_SCHEMA=schema/
    ##SCHEMA="--schema ${MD_SCHEMA} --schema ${WSFED_SCHEMA}"
    #SCHEMA="--schema ${MD_SCHEMA}"
    #

    # We run xmllint against the XML fragment using these
    # schema:
    #   SAML V2.0 metadata namespace (saml-schema-metadata-2.0.xsd)
    #
    my @schemata = qw/
      saml-schema-metadata-2.0.xsd
      ws-federation.xsd
    / ;

    # $schema_dir is where all the .xsd files live.
    my $schema_dir = $SCHEMA_DIR ;

    # Create a place to put the XML catalog.
    my $tmp         = File::Temp->new(UNLINK => 1) ;
    my $xml_catalog = $tmp->filename() ;

    my $xml_catalog_template = File::Spec->catfile(($schema_dir), 'catalog.template');

    local $ENV{'XML_CATALOG_FILES'} = $xml_catalog ;

    make_xml_catalog($schema_dir, $xml_catalog_template, $xml_catalog) ;

    my $saml_metadata_xsd = File::Spec->catfile(($schema_dir), 'saml-schema-metadata-2.0.xsd');
    my $ws_fed_xsd        = File::Spec->catfile(($schema_dir), 'ws-federation.xsd');
    my @cmd = (
        'xmllint',
        '--noout',
        '--nonet',
        '--nowarning',
        '--schema',  $saml_metadata_xsd,
        '--schema',  $ws_fed_xsd,
        $sp_metadata_file,
        ) ;

    my ($stdout, $stderr, $exit_value) = run_command(@cmd) ;

    $stderr = trim($stderr) ;

    my $hint ;
    if ($stderr =~ m{nameidformat}ixsm) {
        $hint = q{NameIDFormat must come AFTER all KeyDescriptor's but }
            . "BEFORE any AssertionConsumerService's.\nSee also "
            . 'saml-schema-metadata-2.0.xsd.' ;
    } elsif ($stderr =~ m{contactperson}ixsm) {
        $hint = 'Make sure that ContactPerson has the same namespace '
            . 'prefix as the rest of the elements.' ;
    }

    if ($hint) {
        $stderr = $stderr . "\n\nHint: $hint\n" ;
    }

    return ($stdout, $stderr, $exit_value) ;
}

sub make_xml_catalog {
    my ($schema_dir, $catalog_file_template, $catalog_file) = @_ ;

    my $schema_template = slurp($catalog_file_template) ;

    $schema_template =~ s{%%SCHEMA_DIR%%}{$schema_dir}gxsm ;

    spew($schema_template, $catalog_file) ;

    return ;
}

sub spew {
    my ($x, $filename) = @_ ;

    open(my $FH, '>', $filename)
        or croak "cannot open file '$filename': $ERRNO" ;

    print {$FH} $x
      or croak "print to file '$filename' failed: $ERRNO" ;

    close($FH) ;

    return 1 ;
}

sub slurp {
    my ($filename) = @_ ;

    my $text = q{} ;
    {
        ## no critic (Variables, InputOutput)

        local ($/, *FH) ;

        # Turn off autodie for a bit.
        {
            no autodie qw(open) ;
            if (!open(FH, '<', $filename)) {
                croak "can't open file $filename for slurping: $ERRNO" ;
            }
        }

        $text = <FH> ;
        ## use critic
    }
    return $text ;
}


sub run_command {
    ## no critic(ValuesAndExpressions::ProhibitMagicNumbers)
    my (@cmd) = @_ ;

    my ($in, $out, $err) ;

    run3 \@cmd, \$in, \$out, \$err;
    my $rv = $CHILD_ERROR >> 8 ;
    return ($out, $err, $rv) ;
}

1;


__END__

=head1 NAME

SAML::SPMetadata::Validate - Validate SAML 2.0 SP Metadata

=head1 VERSION

Version 1

=head1 SYNOPSIS

    use SAML::SPMetadata::Validate qw/ validate /  ;

    my @results = validate($xml_string) ;

    my %sp_info = %{ $results[0] } ;
    my @errors  = @{ $results[1] } ;

    if (@errors) {
        print "found errors in your metdata:" ;
        foreach my $error (@errors) {
          print "\n$error" ;
        }
        print "\n" ;
    } else {
        print "Congratulations: no errors in your metdata.\n" ;
        print "We extracted this information:\n" ;
        use Data::Dumper ;
        print Dumper %sp_info ;
    }

=head1 ABSTRACT

A function to validate and extract information from a SAML 2.0 SP
metatadata XML string.

=head1 DESCRIPTION

This packave provides the functions needed to validate and extract
information from a SAML 2.0 SP metatadata XML string. The XML validation
relies heavily on the xmllint executable.

=head1 SUBROUTINES/METHODS

=head2 validate

Takes an SP metadata XML string as input and, if the metdata is valid
returns the array ({ SP metadata values}, []). If there is a validation
problem will return the array (undef, @errors) where @errors is a string
of error messages indicating the problems.

The hash reference { SP metadata values} will contain these values:

  'entityID'       => the entityID
  'contact_person' => the GivenName for the first ContactPerson element
  'contact_email'  => the EmailAddress for the first ContactPerson element

The error messages hash reference contains mappings from keys in
%ERROR_MESSAGES to error messages.

=head2 make_xml_catalog

Creates an XML catalog on the fly. This is needed so that xmllint knows
which XSD (schema) files to load.

=head2 parse_node_certs

Takes the metdata string and its corresponding parsed XML node object as
input and returns an array of certificate strings, one for each
certificate in the metadata. Of course, if the metadata has no
certificates, it returns the empty array.

=head2 parse_node_contact

Takes as input a Contact node and returns a hash with these keys:

    EmailAddress
    GivenName
    contactType

=head2 parse_xml

Takes as input the metadata XML string and returns a hash with these keys:

    entityID
    entityID_domain
    is_urn
    qualifier (e.g., 'md')
    contact_person: a hash
       contactType
       GivenName
       EmailAddress
    certs_aref (an array of certificate PEMs)

=head2 progress

Used to display progress messages if the DEBUG is turned on.

=head2 run_command

Takes an array representing a command to run. Runs this command and
returns stdout, stderr, and the exit value.

=head2 schema_dir

Used to override the default directory where the XML Schema files are kept.

=head2 debug

Used to set the DEBUG variable.

=head2 slurp

Takes as input a file and returns a string of that file's contents.

=head2 spew

Takes as input a string and a filename and creates a file with that
filename and contents the passed in string.

=head2 validate_against_schema

Takes as input a filename and checks that the file is valid SAML 2.0
metadata.

Retruns the standard out, standard error, and exit value of running that
command. If the validation fails the standard error string is appended
with a "Hint" that tries to explain why the validation failed.

=head2 get_certs_from_node

Returns the PEM strings for all certificates in the provided XML node
object.

=head2 get_node_attribute

Given a node and an attribute, return the value of that attribute in the
node. If the node does not contain an attribute with that name, return
undef.

=head1 DIAGNOSTICS

None yet.

=head1 CONFIGURATION AND ENVIRONMENT

The validation process checks the metadata against various XML Schema
files (.xsd). The directory where the XML Schema files are kept can be
changed by calling the schema_dir function.

=head1 DEPENDENCIES

These modules are called:

    Carp
    File::Spec
    File::Copy
    File::Temp
    XML::LibXM
    Text::Wrap
    IPC::Run3
    Crypt::X509
    MIME::Base6

In addition, the xmllint program must be in the PATH.

=head1 INCOMPATIBILITIES

None known.

=head1 BUGS AND LIMITATIONS

None known.

=head1 AUTHOR

Adam H. Lewenberg (adamhl@stanford.edu)

=head1 LICENSE AND COPYRIGHT

Copyright 2016 Stanford University, all rights reserved.
