package SAML::SPMetadata::X509Certificate ;

# We bend the variable naming convention in some cases since we want the
# variable name to match the XML attribute name:
## no critic (NamingConventions::Capitalization)
## no critic (CodeLayout::ProhibitParensWithBuiltins)
## no critic (Modules::RequireNoMatchVarsWithUseEnglish)

# We like "my $self = shift; my (args) = @_;" pattern, so turn off
# perlcritic's complaint atbout RequireArgUnpacking.
## no critic (Subroutines::RequireArgUnpacking)

# Perl object to represent a certificate used in an SP metadata XML
# string.

use strict ;
use warnings ;
use autodie ;

use Carp;
use Crypt::X509 ; # libcrypt-x509-perl
use Crypt::OpenSSL::X509 ; # 2038 year issue
use English;
use Time::Local ;
use MIME::Base64;
use Data::Dumper ;

# Make this a Moose class
use Moose ;
use namespace::autoclean;

has pem => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
) ;

has notBefore_epoch => (
  is        => 'rw',
  isa       => 'Int',
  required  => 1,
    ) ;

has notAfter_epoch => (
  is        => 'rw',
  isa       => 'Int',
  required  => 1,
    ) ;

has subject => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
) ;

has issuer => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
) ;

has fingerprint_md5 => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
) ;

has fingerprint_sha256 => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
) ;

has isSelfSigned => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
    ) ;

# This is really "use"
has usage => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
    ) ;


# From an XML string element X509Certificate construct the object.
# If the certificate string is not parsable, return nothing.
sub from_pem {
    my ($pem) = @_ ;

    # Remove all whitespace from $pem (Crypt::OpenSSL::X509 has problems
    # with PEM files with extra line feeds).
    #
    # We use Crypt::OpenSSL::X509 instead of Crypt::X509 because
    # Crypt::X509 has Y2038 problems.
    $pem =~ s{\s}{}gxsm ;

    my $cert_string = "-----BEGIN CERTIFICATE-----\n" . $pem . "\n" . "-----END CERTIFICATE-----\n";

    my $openssl_cert;
    eval {
        $openssl_cert = Crypt::OpenSSL::X509->new_from_string($cert_string);
    } ;
    my $aterror = $EVAL_ERROR;

    if ($aterror) {
        croak "Error parsing X509 certificate: $aterror";
    }

    my %mon2num = qw(Jan 0 Feb 1 Mar 2 Apr 3 May 4 Jun 5 Jul 6 Aug 7 Sep 8 Oct 9 Nov 10 Dec 11);

    my @ts_before = split(/\s+/xsm, $openssl_cert->notBefore);
    my @ts_after  = split(/\s+/xsm, $openssl_cert->notAfter);

    ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)
    my @hms_b     = split(/:/xsm, $ts_before[2]);
    my $notBefore = timelocal($hms_b[2], $hms_b[1], $hms_b[0], $ts_before[1], $mon2num{$ts_before[0]}, $ts_before[3], 0, 0, 0);

    my @hms_a     = split(/:/xsm, $ts_after[2]);
    my $notAfter  = timelocal($hms_a[2], $hms_a[1], $hms_a[0], $ts_after[1], $mon2num{$ts_after[0]}, $ts_after[3], 0, 0, 0);

    my $subject = join(q{,}, $openssl_cert->subject) ;
    my $issuer  = join(q{,}, $openssl_cert->issuer) ;

    my $fingerprint_md5    = $openssl_cert->fingerprint_md5() ;
    my $fingerprint_sha256 = $openssl_cert->fingerprint_sha256() ;

    my $self_signed ;
    if ($issuer eq $subject) {
        $self_signed = 1 ;
    } else {
        $self_signed = 0 ;
    }

    my $cert = SAML::SPMetadata::X509Certificate->new(
        'pem'                => $pem,
        'subject'            => $subject,
        'issuer'             => $issuer,
        'notBefore_epoch'    => $notBefore,
        'notAfter_epoch'     => $notAfter,
        'isSelfSigned'       => $self_signed,
        'fingerprint_md5'    => $fingerprint_md5,
        'fingerprint_sha256' => $fingerprint_sha256,
        ) ;

    return $cert ;
}

sub to_hash {
    my $self = shift ;

    my %hash = $self->serialize('hash') ;

    return \%hash ;
}

sub to_string {
    my $self = shift ;
    return $self->serialize('string') ;
}

sub serialize {
    my $self = shift ;
    my ($action) = @_ ;

    my $notBefore_epoch = $self->notBefore_epoch() ;
    my $notAfter_epoch  = $self->notAfter_epoch() ;

    my $subject = $self->subject() ;
    my $issuer  = $self->issuer() ;

    my $selfSigned = $self->isSelfSigned() ;
    my $selfSigned_txt ;
    if ($self->isSelfSigned()) {
        $selfSigned_txt = 1 ;
    } else {
        $selfSigned_txt = 0 ;
    }

    my $fingerprint_sha256 = $self->fingerprint_sha256() ;
    my $fingerprint_md5    = $self->fingerprint_md5() ;

    if ($action eq 'string') {
        return <<"EOS";
subject:            $subject
issuer:             $issuer
not_before:         $notBefore_epoch
not_after:          $notAfter_epoch
fingerprint_md5:    $fingerprint_md5
fingerprint_sha256: $fingerprint_sha256
self signed?:       $selfSigned_txt
EOS

    } elsif ($action eq 'hash') {
        my %hash = () ;
        $hash{'subject'}            = $subject ;
        $hash{'issuer'}             = $issuer ;
        $hash{'notBefore_epoch'}    = $notBefore_epoch ;
        $hash{'notAfter_epoch'}     = $notAfter_epoch ;
        $hash{'selfSigned'}         = $selfSigned ;
        $hash{'fingerprint_md5'}    = $fingerprint_md5 ;
        $hash{'fingerprint_sha256'} = $fingerprint_sha256 ;

        return %hash ;
    }
}


1 ;
